@@ -18,19 +18,13 @@
 #include "mythexp.h"
 #include "mythdate.h"
 #include "mythtypes.h"
+#include "enums/recStatus.h"
 
-/* If NUMPROGRAMLINES gets updated following files need
-   updates and code changes:
-   mythweb/modules/tv/classes/Program.php (layout)
-   mythtv/bindings/perl/MythTV.pm (version number)
-   mythtv/bindings/perl/MythTV/Program.pm (layout)
-   mythtv/bindings/php/MythBackend.php (version number)
-   mythtv/bindings/php/MythTVProgram.php (layout)
-   mythtv/bindings/php/MythTVRecording.php (layout)
-   mythtv/bindings/python/MythTV/static.py (version number)
-   mythtv/bindings/python/MythTV/mythproto.py (layout)
+/* If NUMPROGRAMLINES gets updated, then MYTH_PROTO_VERSION and MYTH_PROTO_TOKEN
+   in mythversion.h need to be bumped, and also follow the instructions in
+   myth_version.h about other changes needed when the network protocol changes.
 */
-#define NUMPROGRAMLINES 47
+#define NUMPROGRAMLINES 52
 
 class ProgramInfo;
 typedef AutoDeleteDeque<ProgramInfo*> ProgramList;
@@ -75,19 +69,23 @@ class MPUBLIC ProgramInfo
   public:
     enum CategoryType { kCategoryNone, kCategoryMovie, kCategorySeries,
                         kCategorySports, kCategoryTVShow };
-                        
+
     /// Null constructor
     ProgramInfo(void);
     /// Copy constructor
     ProgramInfo(const ProgramInfo &other);
     /// Constructs a ProgramInfo from data in 'recorded' table
+    explicit ProgramInfo(uint recordedid);
+    /// Constructs a ProgramInfo from data in 'recorded' table
     ProgramInfo(uint chanid, const QDateTime &recstartts);
     /// Constructs a ProgramInfo from data in 'recorded' table
-    ProgramInfo(const QString &title,
+    ProgramInfo(uint recordedid,
+                const QString &title,
                 const QString &subtitle,
                 const QString &description,
                 uint season,
                 uint episode,
+                uint totalepisodes,
                 const QString &syndicatedepisode,
                 const QString &category,
 
@@ -108,6 +106,7 @@ class MPUBLIC ProgramInfo
                 const QString &seriesid,
                 const QString &programid,
                 const QString &inetref,
+                CategoryType  catType,
 
                 int recpriority,
 
@@ -127,7 +126,7 @@ class MPUBLIC ProgramInfo
                 const QDate &originalAirDate,
                 const QDateTime &lastmodified,
 
-                RecStatusType recstatus,
+                RecStatus::Type recstatus,
 
                 uint recordid,
 
@@ -139,7 +138,9 @@ class MPUBLIC ProgramInfo
                 uint programflags,
                 uint audioproperties,
                 uint videoproperties,
-                uint subtitleType);
+                uint subtitleType,
+                const QString &inputname,
+                const QDateTime &bookmarkupdate);
 
     /// Constructs a ProgramInfo from data in 'oldrecorded' table
     ProgramInfo(const QString &title,
@@ -163,7 +164,7 @@ class MPUBLIC ProgramInfo
                 const QDateTime &recstartts,
                 const QDateTime &recendts,
 
-                RecStatusType recstatus,
+                RecStatus::Type recstatus,
 
                 uint recordid,
 
@@ -200,7 +201,7 @@ class MPUBLIC ProgramInfo
                 uint partnumber,
                 uint parttotal,
                 const QDate &originalAirDate,
-                RecStatusType recstatus,
+                RecStatus::Type recstatus,
                 uint recordid,
                 RecordingType rectype,
                 uint findid,
@@ -212,6 +213,10 @@ class MPUBLIC ProgramInfo
                 uint audioprops,
                 uint subtitletype,
 
+                uint season,
+                uint episode,
+                uint totalepisodes,
+
                 const ProgramList &schedList);
     /// Constructs a basic ProgramInfo (used by RecordingInfo)
     ProgramInfo(const QString &title,
@@ -219,6 +224,7 @@ class MPUBLIC ProgramInfo
                 const QString &description,
                 uint season,
                 uint episode,
+                uint totalepisodes,
                 const QString &category,
 
                 uint chanid,
@@ -239,7 +245,7 @@ class MPUBLIC ProgramInfo
                 const QString &programid,
                 const QString &inetref);
     /// Constructs a ProgramInfo for a pathname.
-    ProgramInfo(const QString &pathname);
+    explicit ProgramInfo(const QString &pathname);
     /// Constructs a ProgramInfo for a video.
     ProgramInfo(const QString &pathname,
                 const QString &plot,
@@ -265,7 +271,7 @@ class MPUBLIC ProgramInfo
         if (!FromStringList(it, end))
             clear();
     }
-    ProgramInfo(const QStringList &list) :
+    explicit ProgramInfo(const QStringList &list) :
         chanid(0),
         positionMapDBReplacement(NULL)
     {
@@ -291,9 +297,11 @@ class MPUBLIC ProgramInfo
     virtual void SubstituteMatches(QString &str);
 
     // Used for scheduling recordings
-    bool IsSameProgram(const ProgramInfo &other) const;
-    bool IsSameTimeslot(const ProgramInfo &other) const;
-    bool IsSameProgramTimeslot(const ProgramInfo &other) const;//sched only
+    bool IsSameProgram(const ProgramInfo &other) const; // Exact same program
+    bool IsDuplicateProgram(const ProgramInfo &other) const; // Is this program considered a duplicate according to rule type and dup method (scheduler only)
+    bool IsSameProgramAndStartTime(const ProgramInfo &other) const; // Exact same program and same starttime, Any channel
+    bool IsSameTitleStartTimeAndChannel(const ProgramInfo &other) const; // Same title, starttime and channel
+    bool IsSameTitleTimeslotAndChannel(const ProgramInfo &other) const;//sched only - Same title, starttime, endtime and channel
     static bool UsingProgramIDAuthority(void)
     {
         return usingProgIDAuth;
@@ -304,6 +312,7 @@ class MPUBLIC ProgramInfo
     bool IsSameProgramWeakCheck(const ProgramInfo &other) const;
     bool IsSameRecording(const ProgramInfo &other) const
         { return chanid == other.chanid && recstartts == other.recstartts; }
+    bool IsSameChannel(const ProgramInfo &other) const;
 
     // Quick gets
     /// Creates a unique string that can be used to identify an
@@ -335,6 +344,7 @@ class MPUBLIC ProgramInfo
     QString GetDescription(void)  const { return description; }
     uint    GetSeason(void)       const { return season; }
     uint    GetEpisode(void)      const { return episode; }
+    uint    GetEpisodeTotal(void) const { return totalepisodes; }
     QString GetCategory(void)     const { return category; }
     /// This is the unique key used in the database to locate tuning
     /// information. [1..2^32] are valid keys, 0 is not.
@@ -400,7 +410,7 @@ class MPUBLIC ProgramInfo
     {
         return MythDate::toString(lastmodified, fmt);
     }
-    uint64_t GetFilesize(void)            const { return filesize;     }
+    virtual uint64_t GetFilesize(void)    const; // TODO Remove
     QString GetSeriesID(void)             const { return seriesid;     }
     QString GetProgramID(void)            const { return programid;    }
     QString GetInetRef(void)              const { return inetref;      }
@@ -412,10 +422,9 @@ class MPUBLIC ProgramInfo
     uint    GetStars(uint range_max)      const
         { return (int)(stars * range_max + 0.5f); }
 
-    RecStatusType GetRecordingStatus(void)    const
-        { return (RecStatusType)recstatus; }
-    RecStatusType GetOldRecordingStatus(void) const
-        { return (RecStatusType)oldrecstatus; }
+    uint    GetRecordingID(void)              const { return recordedid; }
+    RecStatus::Type GetRecordingStatus(void)    const
+        { return (RecStatus::Type)recstatus; }
     uint    GetPreferedInputID(void)          const { return prefinput; }
     uint    GetRecordingRuleID(void)          const { return recordid;  }
     uint    GetParentRecordingRuleID(void)    const { return parentid;  }
@@ -432,13 +441,14 @@ class MPUBLIC ProgramInfo
 
     uint    GetSourceID(void)             const { return sourceid;     }
     uint    GetInputID(void)              const { return inputid;      }
-    uint    GetCardID(void)               const { return cardid;       }
+    QString GetInputName(void)            const { return inputname;    }
 
     uint    GetFindID(void)               const { return findid;       }
 
     uint32_t GetProgramFlags(void)        const { return programflags; }
     ProgramInfoType GetProgramInfoType(void) const
-        { return (ProgramInfoType)((programflags & FL_TYPEMASK) >> 16); }
+        { return (ProgramInfoType)((programflags & FL_TYPEMASK) >> 20); }
+    QDateTime GetBookmarkUpdate(void) const { return bookmarkupdate; }
     bool IsGeneric(void) const;
     bool IsInUsePlaying(void)   const { return programflags & FL_INUSEPLAYING;}
     bool IsCommercialFree(void) const { return programflags & FL_CHANCOMMFREE;}
@@ -475,7 +485,7 @@ class MPUBLIC ProgramInfo
     // Quick sets
     void SetTitle(const QString &t) { title = t; title.detach(); }
     void SetProgramInfoType(ProgramInfoType t)
-        { programflags &= ~FL_TYPEMASK; programflags |= ((uint32_t)t<<16); }
+        { programflags &= ~FL_TYPEMASK; programflags |= ((uint32_t)t<<20); }
     void SetPathname(const QString&) const;
     void SetChanID(uint _chanid) { chanid = _chanid; }
     void SetScheduledStartTime(const QDateTime &dt) { startts      = dt;    }
@@ -486,7 +496,7 @@ class MPUBLIC ProgramInfo
     void SetPlaybackGroup( const QString &group)    { playgroup    = group; }
     void SetHostname(      const QString &host)     { hostname     = host;  }
     void SetStorageGroup(  const QString &group)    { storagegroup = group; }
-    void SetFilesize(      uint64_t       sz)       { filesize     = sz;    }
+    virtual void SetFilesize( uint64_t       sz); /// TODO Move to RecordingInfo
     void SetSeriesID(      const QString &id)       { seriesid     = id;    }
     void SetProgramID(     const QString &id)       { programid    = id;    }
     void SetCategory(      const QString &cat)      { category     = cat;   }
@@ -495,8 +505,7 @@ class MPUBLIC ProgramInfo
     void SetRecordingPriority2(int priority)     { recpriority2 = priority; }
     void SetRecordingRuleID(uint id)                { recordid     = id;    }
     void SetSourceID(uint id)                       { sourceid     = id;    }
-    void SetInputID( uint id)                       { inputid      = id;    }
-    void SetCardID(  uint id)                       { cardid       = id;    }
+    void SetInputID(uint id)                        { inputid      = id;    }
     void SetReactivated(bool reactivate)
     {
         programflags &= ~FL_REACTIVATE;
@@ -519,19 +528,38 @@ class MPUBLIC ProgramInfo
         programflags &= ~FL_IGNOREBOOKMARK;
         programflags |= (ignore) ? FL_IGNOREBOOKMARK : 0;
     }
-    void SetRecordingStatus(RecStatusType status) { recstatus = status; }
+    /// \brief If "ignore" is true QueryProgStart() will return 0, otherwise
+    ///        QueryProgStart() will return the progstart position if it exists.
+    void SetIgnoreProgStart(bool ignore)
+    {
+        programflags &= ~FL_IGNOREPROGSTART;
+        programflags |= (ignore) ? FL_IGNOREPROGSTART : 0;
+    }
+    /// \brief If "ignore" is true QueryLastPlayPos() will return 0, otherwise
+    ///        QueryLastPlayPos() will return the last playback position
+    ///        if it exists.
+    void SetAllowLastPlayPos(bool allow)
+    {
+        programflags &= ~FL_ALLOWLASTPLAYPOS;
+        programflags |= (allow) ? FL_ALLOWLASTPLAYPOS : 0;
+    }
+    virtual void SetRecordingID(uint _recordedid) { recordedid = _recordedid; }
+    void SetRecordingStatus(RecStatus::Type status) { recstatus = status; }
     void SetRecordingRuleType(RecordingType type) { rectype   = type;   }
     void SetPositionMapDBReplacement(PMapDBReplacement *pmap)
         { positionMapDBReplacement = pmap; }
 
     // Slow DB gets
     QString     QueryBasename(void) const;
-    uint64_t    QueryFilesize(void) const;
+//  uint64_t    QueryFilesize(void) const; // TODO Remove
     uint        QueryMplexID(void) const;
     QDateTime   QueryBookmarkTimeStamp(void) const;
     uint64_t    QueryBookmark(void) const;
+    uint64_t    QueryProgStart(void) const;
+    uint64_t    QueryLastPlayPos(void) const;
     CategoryType QueryCategoryType(void) const;
     QStringList QueryDVDBookmark(const QString &serialid) const;
+    QStringList QueryBDBookmark(const QString &serialid) const;
     bool        QueryIsEditing(void) const;
     bool        QueryIsInUse(QStringList &byWho) const;
     bool        QueryIsInUse(QString &byWho) const;
@@ -544,7 +572,7 @@ class MPUBLIC ProgramInfo
     uint        QueryAverageHeight(void) const;
     uint        QueryAverageFrameRate(void) const;
     MarkTypes   QueryAverageAspectRatio(void) const;
-    int64_t     QueryTotalDuration(void) const;
+    uint32_t    QueryTotalDuration(void) const;
     int64_t     QueryTotalFrames(void) const;
     QString     QueryRecordingGroup(void) const;
     bool        QueryMarkupFlag(MarkTypes type) const;
@@ -553,9 +581,10 @@ class MPUBLIC ProgramInfo
     bool        Reload(void);
 
     // Slow DB sets
-    void SaveFilesize(uint64_t fsize);
+    virtual void SaveFilesize(uint64_t fsize); /// TODO Move to RecordingInfo
     void SaveBookmark(uint64_t frame);
     void SaveDVDBookmark(const QStringList &fields) const;
+    void SaveBDBookmark(const QStringList &fields) const;
     void SaveEditing(bool edit);
     void SaveTranscodeStatus(TranscodingStatus transFlag);
     void SaveWatched(bool watchedFlag);
@@ -599,6 +628,22 @@ class MPUBLIC ProgramInfo
                          int64_t min_frm = -1, int64_t max_frm = -1) const;
     void SavePositionMapDelta(frm_pos_map_t &, MarkTypes type) const;
 
+    // Get position/duration for keyframe and vice versa
+    bool QueryKeyFrameInfo(uint64_t *, uint64_t position_or_keyframe,
+                           bool backwards,
+                           MarkTypes type, const char *from_filemarkup_asc,
+                           const char *from_filemarkup_desc,
+                           const char *from_recordedseek_asc,
+                           const char *from_recordedseek_desc) const;
+    bool QueryKeyFramePosition(uint64_t *, uint64_t keyframe,
+                               bool backwards) const;
+    bool QueryPositionKeyFrame(uint64_t *, uint64_t position,
+                               bool backwards) const;
+    bool QueryKeyFrameDuration(uint64_t *, uint64_t keyframe,
+                               bool backwards) const;
+    bool QueryDurationKeyFrame(uint64_t *, uint64_t duration,
+                               bool backwards) const;
+
     // Get/set all markup
     struct MarkupEntry
     {
@@ -633,6 +678,8 @@ class MPUBLIC ProgramInfo
         const QString &pathname, uint &chanid, QDateTime &recstartts);
     static bool QueryKeyFromPathname(
         const QString &pathname, uint &chanid, QDateTime &recstartts);
+    static bool QueryRecordedIdFromPathname(const QString &pathname,
+        uint &recordedid);
 
     static QString  QueryRecordingGroupPassword(const QString &group);
     static uint64_t QueryBookmark(uint chanid, const QDateTime &recstartts);
@@ -640,7 +687,6 @@ class MPUBLIC ProgramInfo
     static QMap<QString,bool> QueryJobsRunning(int type);
     static QStringList LoadFromScheduler(const QString &altTable, int recordid);
 
-  protected:
     // Flagging map support methods
     void QueryMarkupMap(frm_dir_map_t&, MarkTypes type,
                         bool merge = false) const;
@@ -649,6 +695,7 @@ class MPUBLIC ProgramInfo
     void ClearMarkupMap(MarkTypes type = MARK_ALL,
                         int64_t min_frm = -1, int64_t max_frm = -1) const;
 
+  protected:
     // Creates a basename from the start and end times
     QString CreateRecordBasename(const QString &ext) const;
 
@@ -673,6 +720,7 @@ class MPUBLIC ProgramInfo
     QString description;
     uint    season;
     uint    episode;
+    uint    totalepisodes;
     QString syndicatedepisode;
     QString category;
     QString director;
@@ -680,8 +728,8 @@ class MPUBLIC ProgramInfo
     int32_t recpriority;
 
     uint32_t chanid;
-    QString chanstr;
-    QString chansign;
+    QString chanstr; // Channum
+    QString chansign; // Callsign
     QString channame;
     QString chanplaybackfilters;
 
@@ -718,7 +766,6 @@ class MPUBLIC ProgramInfo
 
     uint32_t sourceid;
     uint32_t inputid;
-    uint32_t cardid;
     uint32_t findid;
 
     uint32_t programflags;    ///< ProgramFlag
@@ -728,11 +775,14 @@ class MPUBLIC ProgramInfo
     uint16_t parttotal;
 
     int8_t recstatus;
-    int8_t oldrecstatus;
     uint8_t rectype;
     uint8_t dupin;
     uint8_t dupmethod;
 
+    uint    recordedid;
+    QString inputname;
+    QDateTime bookmarkupdate;
+
 // everything below this line is not serialized
     uint8_t availableStatus; // only used for playbackbox.cpp
   public:
@@ -754,6 +804,13 @@ class MPUBLIC ProgramInfo
     static bool usingProgIDAuth;
 };
 
+MPUBLIC bool LoadFromProgram(
+    ProgramList        &destination,
+    const QString      &where,
+    const QString      &groupBy,
+    const QString      &orderBy,
+    const MSqlBindings &bindings,
+    const ProgramList  &schedList);
 
 MPUBLIC bool LoadFromProgram(
     ProgramList        &destination,
@@ -761,11 +818,31 @@ MPUBLIC bool LoadFromProgram(
     const MSqlBindings &bindings,
     const ProgramList  &schedList);
 
+MPUBLIC bool LoadFromProgram(
+    ProgramList        &destination,
+    const QString      &sql,
+    const MSqlBindings &bindings,
+    const ProgramList  &schedList,
+    const uint         &start,
+    const uint         &limit,
+    uint               &count);
+
+MPUBLIC ProgramInfo*  LoadProgramFromProgram(
+        const uint chanid, const QDateTime &starttime);
+
 MPUBLIC bool LoadFromOldRecorded(
     ProgramList        &destination,
     const QString      &sql,
     const MSqlBindings &bindings);
 
+MPUBLIC bool LoadFromOldRecorded(
+    ProgramList        &destination, 
+    const QString      &sql, 
+    const MSqlBindings &bindings,
+    const uint         &start, 
+    const uint         &limit,
+    uint               &count);
+
 MPUBLIC bool LoadFromRecorded(
     ProgramList        &destination,
     bool                possiblyInProgressRecordingsOnly,
@@ -795,8 +872,10 @@ bool LoadFromScheduler(
     {
         TYPE *p = new TYPE(sit, slist.end());
         destination.push_back(p);
+
         if (!p->HasPathname() && !p->GetChanID())
         {
+            delete p;
             destination.clear();
             return false;
         }
@@ -822,7 +901,7 @@ bool LoadFromScheduler(AutoDeleteDeque<T> &destination)
 // could be factored out
 MPUBLIC bool GetNextRecordingList(QDateTime &nextRecordingStart,
                                   bool *hasConflicts = NULL,
-                                  vector<ProgramInfo> *list = NULL);
+                                  std::vector<ProgramInfo> *list = NULL);
 
 class QMutex;
 class MPUBLIC PMapDBReplacement
